package cn.minbb.evaluation.ctrl;

import android.app.Dialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.net.Uri;
import android.support.v7.app.AlertDialog;
import android.view.LayoutInflater;
import android.view.View;
import android.widget.TextView;
import android.widget.Toast;

import org.wlf.filedownloader.DownloadFileInfo;
import org.wlf.filedownloader.FileDownloader;
import org.wlf.filedownloader.listener.OnDeleteDownloadFileListener;
import org.wlf.filedownloader.listener.OnFileDownloadStatusListener;
import org.wlf.filedownloader.listener.OnRenameDownloadFileListener;
import org.wlf.filedownloader.listener.simple.OnSimpleFileDownloadStatusListener;

import java.io.File;
import java.text.NumberFormat;

import cn.minbb.evaluation.R;
import cn.minbb.evaluation.views.HorizontalNumberProgressBar;
import cn.minbb.evaluation.entity.Version;

public class UpdateManager {

    private Context context;
    private Version version;

    private Dialog downloadDialog;
    private HorizontalNumberProgressBar progressBar;
    private TextView progressTextView, timeTextView, speedTextView;

    public UpdateManager(Context context, Version version) {
        this.context = context;
        this.version = version;
    }

    /**
     * 检查软件是否需要更新
     */
    public void checkUpdate() {
        if (version != null) {
            // 判断是否忽略此版本的更新
            if (VersionManager.getIgnoreVersionCode() != version.getVersionCode()) {
                // 判断版本号是否最新
                try {
                    int versionCodeLocal = context.getPackageManager().getPackageInfo(context.getPackageName(), 0).versionCode;
                    if (version.getVersionCode() > versionCodeLocal) {
                        // 需要更新，显示提示更新对话框
                        showNoticeDialog();
                    }
                } catch (PackageManager.NameNotFoundException e) {
                    e.printStackTrace();
                }
            }
        } else {
            Toast.makeText(context, "检查更新失败，请检查互联网连接！", Toast.LENGTH_SHORT).show();
        }
    }

    /**
     * 有更新时显示提示对话框
     */
    public void showNoticeDialog() {
        AlertDialog.Builder builder = new AlertDialog.Builder(context);
        builder.setTitle("发现新版本");
        NumberFormat numberFormat = NumberFormat.getNumberInstance();
        numberFormat.setMaximumFractionDigits(2);
        String message = "文件名：" + version.getFileName() + "\n"
                + "大小：" + numberFormat.format(version.getFileSize() / 1024.0 / 1024.0) + " MB\n"
                + version.getContent();
        builder.setMessage(message);
        builder.setCancelable(false);

        builder.setPositiveButton("现在更新", (dialog, which) -> {
            // 隐藏当前对话框
            dialog.dismiss();
            // 文件存在则安装，不存在则下载
            File apkFile = new File(App.getAppDirectoryPath(), version.getFileName());
            if (apkFile.exists()) {
                installAPK(apkFile.getPath());
            } else {
                // 显示下载对话框
                showDownloadDialog();
            }
        });
        builder.setNegativeButton("下次再说", (dialog, which) -> dialog.dismiss());
        // 判断是否强制更新
        if (!version.getForced()) {
            builder.setNeutralButton("忽略此版本", (dialog, i) -> {
                dialog.dismiss();
                // 记录此版本号
                VersionManager.setIgnoreVersionCode(version.getVersionCode());
                Toast.makeText(context, "已忽略此版本更新", Toast.LENGTH_SHORT).show();
            });
        }
        builder.create().show();
    }

    /**
     * 显示正在下载对话框
     */
    private void showDownloadDialog() {
        AlertDialog.Builder builder = new AlertDialog.Builder(context);
        builder.setTitle("下载中");
        builder.setCancelable(false);
        View view = LayoutInflater.from(context).inflate(R.layout.dialog_download_apk, null);
        progressBar = view.findViewById(R.id.progressBar);
        progressTextView = view.findViewById(R.id.progressTextView);
        timeTextView = view.findViewById(R.id.timeTextView);
        speedTextView = view.findViewById(R.id.speedTextView);
        builder.setView(view);
        builder.setPositiveButton("后台下载", new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialog, int which) {
                dialog.dismiss();
            }
        });
        builder.setNegativeButton("取消", new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialog, int which) {
                // 取消下载并删除文件
                FileDownloader.delete(version.getFileUrl(), true, new OnDeleteDownloadFileListener() {
                    @Override
                    public void onDeleteDownloadFilePrepared(DownloadFileInfo downloadFileNeedDelete) {
                    }

                    @Override
                    public void onDeleteDownloadFileSuccess(DownloadFileInfo downloadFileDeleted) {
                        dialog.dismiss();
                    }

                    @Override
                    public void onDeleteDownloadFileFailed(DownloadFileInfo downloadFileInfo, DeleteDownloadFileFailReason failReason) {
                        dialog.dismiss();
                    }
                });
            }
        });
        downloadDialog = builder.create();
        downloadDialog.show();

        // 下载文件
        downloadAPK();
    }

    /**
     * 下载文件
     */
    private void downloadAPK() {
        // 添加【文件下载状态监听器】
        FileDownloader.registerDownloadStatusListener(onFileDownloadStatusListener);
        // 下载文件
        FileDownloader.start(version.getFileUrl());
    }

    /**
     * 下载到本地后执行安装
     *
     * @param filePath 文件路径
     */
    private void installAPK(String filePath) {
        File apkFile = new File(filePath);
        if (!apkFile.exists()) {
            return;
        }
        Intent intent = new Intent(Intent.ACTION_VIEW);
        Uri uri = Uri.parse("file://" + apkFile.toString());
        intent.setDataAndType(uri, "application/vnd.android.package-archive");
        context.startActivity(intent);
    }

    private OnFileDownloadStatusListener onFileDownloadStatusListener = new OnSimpleFileDownloadStatusListener() {
        @Override
        public void onFileDownloadStatusRetrying(DownloadFileInfo downloadFileInfo, int retryTimes) {
            // retrying download when failed once, the retryTimes is the current trying times
        }

        @Override
        public void onFileDownloadStatusWaiting(DownloadFileInfo downloadFileInfo) {
            // waiting for download(wait for other tasks paused, or FileDownloader is busy for other operations)
        }

        @Override
        public void onFileDownloadStatusPreparing(DownloadFileInfo downloadFileInfo) {
            // preparing(connecting)
        }

        @Override
        public void onFileDownloadStatusPrepared(DownloadFileInfo downloadFileInfo) {
            // prepared(connected)
        }

        @Override
        public void onFileDownloadStatusDownloading(DownloadFileInfo downloadFileInfo, float downloadSpeed, long remainingTime) {
            // downloading, the downloadSpeed with KB/s unit, the remainingTime with seconds unit
            progressBar.setProgress((int) (100.0 * downloadFileInfo.getDownloadedSizeLong() / downloadFileInfo.getFileSizeLong()));
            NumberFormat numberFormat = NumberFormat.getNumberInstance();
            numberFormat.setMaximumFractionDigits(2);
            progressTextView.setText(numberFormat.format(downloadFileInfo.getDownloadedSizeLong() / 1048576.0) + "MB / " + numberFormat.format(downloadFileInfo.getFileSizeLong() / 1048576.0) + "MB");
            timeTextView.setText(remainingTime + "s");
            speedTextView.setText(numberFormat.format(downloadSpeed) + "KB/s");
        }

        @Override
        public void onFileDownloadStatusPaused(DownloadFileInfo downloadFileInfo) {
            // download paused
            FileDownloader.unregisterDownloadStatusListener(onFileDownloadStatusListener);
            downloadDialog.dismiss();
        }

        @Override
        public void onFileDownloadStatusCompleted(DownloadFileInfo downloadFileInfo) {
            // download completed(the url file has been finished)
            FileDownloader.unregisterDownloadStatusListener(onFileDownloadStatusListener);
            downloadDialog.dismiss();
            Toast.makeText(context, "下载成功", Toast.LENGTH_SHORT).show();
            // 重命名 APK 文件，然后安装 APK 文件
            FileDownloader.rename(version.getFileUrl(), version.getFileName(), true, new OnRenameDownloadFileListener() {
                @Override
                public void onRenameDownloadFilePrepared(DownloadFileInfo downloadFileNeedRename) {
                }

                @Override
                public void onRenameDownloadFileSuccess(DownloadFileInfo downloadFileRenamed) {
                    installAPK(downloadFileInfo.getFilePath());
                }

                @Override
                public void onRenameDownloadFileFailed(DownloadFileInfo downloadFileInfo, RenameDownloadFileFailReason failReason) {
                    FileDownloader.delete(version.getFileUrl(), true, null);
                }
            });
        }

        @Override
        public void onFileDownloadStatusFailed(String url, DownloadFileInfo downloadFileInfo, FileDownloadStatusFailReason failReason) {
            // error occur, see failReason for details, some of the failReason you must concern
            FileDownloader.unregisterDownloadStatusListener(onFileDownloadStatusListener);
            downloadDialog.dismiss();

            String failType = failReason.getType();
            String failUrl = failReason.getUrl();// or failUrl = url, both url and failReason.getUrl() are the same

            if (FileDownloadStatusFailReason.TYPE_URL_ILLEGAL.equals(failType)) {
                // the url error when downloading file with failUrl
            } else if (FileDownloadStatusFailReason.TYPE_STORAGE_SPACE_IS_FULL.equals(failType)) {
                // storage space is full when downloading file with failUrl
            } else if (FileDownloadStatusFailReason.TYPE_NETWORK_DENIED.equals(failType)) {
                // network access denied when downloading file with failUrl
            } else if (FileDownloadStatusFailReason.TYPE_NETWORK_TIMEOUT.equals(failType)) {
                // connect timeout when downloading file with failUrl
            } else {
                // more....
            }

            // exception details
            Throwable failCause = failReason.getCause();// or failReason.getOriginalCause()

            // also you can see the exception message
            String failMsg = failReason.getMessage();// or failReason.getOriginalCause().getMessage()
        }
    };
}
