package cn.minbb.evaluation.utils;

import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;

public class BitmapUtil {

    private static int mDesiredWidth;
    private static int mDesiredHeight;

    /**
     * 从 Resources 中加载图片
     *
     * @param res
     * @param resId
     * @param reqWidth
     * @param reqHeight
     * @return
     */
    public static Bitmap decodeSampledBitmapFromResource(Resources res, int resId, int reqWidth, int reqHeight) {
        BitmapFactory.Options options = new BitmapFactory.Options();
        // 设置成了 true,不占用内存，只获取 bitmap 宽高
        options.inJustDecodeBounds = true;
        // 初始化 options 对象
        BitmapFactory.decodeResource(res, resId, options);
        // 得到计算好的 options，目标宽、目标高
        options = getBestOptions(options, reqWidth, reqHeight);
        // 载入一个稍大的缩略图
        Bitmap src = BitmapFactory.decodeResource(res, resId, options);
        // 进一步得到目标大小的缩略图
        return createScaleBitmap(src, mDesiredWidth, mDesiredHeight);
    }

    /**
     * 从 SD 卡上加载图片
     *
     * @param pathName
     * @param reqWidth
     * @param reqHeight
     * @return
     */
    public static Bitmap decodeSampledBitmapFromFile(String pathName, int reqWidth, int reqHeight) {
        BitmapFactory.Options options = new BitmapFactory.Options();
        options.inJustDecodeBounds = true;
        BitmapFactory.decodeFile(pathName, options);
        options = getBestOptions(options, reqWidth, reqHeight);
        Bitmap src = BitmapFactory.decodeFile(pathName, options);
        return createScaleBitmap(src, mDesiredWidth, mDesiredHeight);
    }

    /**
     * 通过传入的 bitmap，进行压缩，得到符合标准的 bitmap
     *
     * @param tempBitmap    原位图
     * @param desiredWidth  位图的期望宽度
     * @param desiredHeight 位图的期望高度
     * @return 压缩后的位图
     */
    public static Bitmap createScaleBitmap(Bitmap tempBitmap, int desiredWidth, int desiredHeight) {
        // 必要时，缩小到最大可接受尺寸。
        if (tempBitmap != null && (tempBitmap.getWidth() > desiredWidth || tempBitmap.getHeight() > desiredHeight)) {
            // 如果是放大图片，filter 决定是否平滑，如果是缩小图片，filter 无影响
            Bitmap bitmap = Bitmap.createScaledBitmap(tempBitmap, desiredWidth, desiredHeight, true);
            // 释放 Bitmap 的 native 像素数组
            tempBitmap.recycle();
            return bitmap;
        } else {
            // 如果没有缩放，那么不回收
            return tempBitmap;
        }
    }

    /**
     * 计算目标宽度，目标高度，inSampleSize
     *
     * @param options
     * @param reqWidth
     * @param reqHeight
     * @return BitmapFactory.Options 对象
     */
    private static BitmapFactory.Options getBestOptions(BitmapFactory.Options options, int reqWidth, int reqHeight) {
        // 读取图片长宽
        int actualWidth = options.outWidth;
        int actualHeight = options.outHeight;
        // 然后计算我们最理想的解码尺寸。
        mDesiredWidth = getResizedDimension(reqWidth, reqHeight, actualWidth, actualHeight);
        mDesiredHeight = getResizedDimension(reqHeight, reqWidth, actualHeight, actualWidth);
        // 根据现在得到计算 inSampleSize
        options.inSampleSize = calculateBestInSampleSize(actualWidth, actualHeight, mDesiredWidth, mDesiredHeight);
        // 使用获取到的inSampleSize 值再次解析图片
        options.inJustDecodeBounds = false;
        return options;
    }

    /**
     * 缩放矩形的一边以适应宽高比。最终得到重新测量的尺寸。
     *
     * @param maxPrimary      最大尺寸的最大尺寸（即最大宽度）宽度，或零以保持二次维度的长宽比
     * @param maxSecondary    次要尺寸的最大尺寸，或零保持基本尺寸的长宽比
     * @param actualPrimary   实际大小的主要维度
     * @param actualSecondary 实际大小的次要维度
     */
    private static int getResizedDimension(int maxPrimary, int maxSecondary, int actualPrimary, int actualSecondary) {
        double ratio = (double) actualSecondary / (double) actualPrimary;
        int resized = maxPrimary;
        if (resized * ratio > maxSecondary) {
            resized = (int) (maxSecondary / ratio);
        }
        return resized;
    }

    /**
     * 返回两个除数的最大功率，用于缩小位图，它不会导致缩放到所需的维度。
     *
     * @param actualWidth   位图的实际宽度
     * @param actualHeight  位图的实际高度
     * @param desiredWidth  位图的期望宽度
     * @param desiredHeight 位图的期望高度
     */
    private static int calculateBestInSampleSize(int actualWidth, int actualHeight, int desiredWidth, int desiredHeight) {
        double wr = (double) actualWidth / desiredWidth;
        double hr = (double) actualHeight / desiredHeight;
        double ratio = Math.min(wr, hr);
        float inSampleSize = 1.0f;
        while ((inSampleSize * 2) <= ratio) {
            inSampleSize *= 2;
        }
        return (int) inSampleSize;
    }
}
