package cn.minbb.evaluation.views;

import android.content.Context;
import android.content.res.TypedArray;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.util.AttributeSet;
import android.util.TypedValue;
import android.widget.ProgressBar;

import cn.minbb.evaluation.R;

public class HorizontalNumberProgressBar extends ProgressBar {

    private static final int DEFAULT_TEXT_SIZE = 12;
    private static final int DEFAULT_TEXT_COLOR = 0xFFFF4081;
    private static final int DEFAULT_TEXT_OFFSET = 8;
    private static final int DEFAULT_REACHED_COLOR = DEFAULT_TEXT_COLOR;
    private static final int DEFAULT_REACHED_HEIGHT = 3;
    private static final int DEFAULT_UNREACHED_COLOR = 0xFFD3D6DA;
    private static final int DEFAULT_UNREACHED_HEIGHT = 3;

    private int textSize = sp2px(DEFAULT_TEXT_SIZE);
    private int textColor = DEFAULT_TEXT_COLOR;
    private int textOffset = dp2px(DEFAULT_TEXT_OFFSET);
    private int reachedColor = DEFAULT_REACHED_COLOR;
    private int reachedHeight = DEFAULT_REACHED_HEIGHT;
    private int unreachedColor = DEFAULT_UNREACHED_COLOR;
    private int unreachedHeight = DEFAULT_UNREACHED_HEIGHT;

    private Paint paint = new Paint();

    private int realWidth;

    public HorizontalNumberProgressBar(Context context) {
        this(context, null);
    }

    public HorizontalNumberProgressBar(Context context, AttributeSet attrs) {
        this(context, attrs, 0);

    }

    public HorizontalNumberProgressBar(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        // 获取自定义属性
        TypedArray typedArray = context.obtainStyledAttributes(attrs, R.styleable.HorizontalNumberProgressBar);
        textSize = (int) typedArray.getDimension(R.styleable.HorizontalNumberProgressBar_progress_text_size, textSize);
        textColor = typedArray.getColor(R.styleable.HorizontalNumberProgressBar_progress_text_color, textColor);
        textOffset = (int) typedArray.getDimension(R.styleable.HorizontalNumberProgressBar_progress_text_offset, textOffset);
        reachedColor = typedArray.getColor(R.styleable.HorizontalNumberProgressBar_progress_reached_color, reachedColor);
        reachedHeight = (int) typedArray.getDimension(R.styleable.HorizontalNumberProgressBar_progress_reached_height, reachedHeight);
        unreachedColor = typedArray.getColor(R.styleable.HorizontalNumberProgressBar_progress_unreached_color, unreachedColor);
        unreachedHeight = (int) typedArray.getDimension(R.styleable.HorizontalNumberProgressBar_progress_unreached_height, unreachedHeight);
        typedArray.recycle();
        paint.setTextSize(textSize);
    }

    @Override
    protected synchronized void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        // int widthMode = MeasureSpec.getMode(widthMeasureSpec);
        int widthValue = MeasureSpec.getSize(widthMeasureSpec);
        int height = measureHeight(heightMeasureSpec);

        setMeasuredDimension(widthValue, height);

        realWidth = getMeasuredWidth() - getPaddingLeft() - getPaddingRight();
    }

    @Override
    protected synchronized void onDraw(Canvas canvas) {
        canvas.save();
        canvas.translate(getPaddingLeft(), getHeight() / 2);

        boolean noNeedUnreached = false;

        // 绘制到达进度
        String text = getProgress() + "%";
        int textWidth = (int) paint.measureText(text);
        float radio = getProgress() * 1.0f / getMax();
        float progressX = radio * realWidth;
        if (progressX + textWidth > realWidth) {
            progressX = realWidth - textWidth;
            noNeedUnreached = true;
        }
        float endX = radio * realWidth - textOffset / 2;
        if (endX > 0) {
            paint.setColor(reachedColor);
            paint.setStrokeWidth(reachedHeight);
            canvas.drawLine(0, 0, endX, 0, paint);
        }
        // 绘制文本
        paint.setColor(textColor);
        int y = (int) (-(paint.descent() + paint.ascent()) / 2);
        canvas.drawText(text, progressX, y, paint);
        // 绘制未到达进度
        if (!noNeedUnreached) {
            float start = progressX + textOffset / 2 + textWidth;
            paint.setColor(unreachedColor);
            paint.setStrokeWidth(unreachedHeight);
            canvas.drawLine(start, 0, realWidth, 0, paint);
        }

        canvas.restore();
    }

    private int measureHeight(int heightMeasureSpec) {
        int result = 0;
        int mode = MeasureSpec.getMode(heightMeasureSpec);
        int size = MeasureSpec.getSize(heightMeasureSpec);

        if (mode == MeasureSpec.EXACTLY) {
            result = size;
        } else {
            int textHeight = (int) (paint.descent() - paint.ascent());
            result = getPaddingTop() + getPaddingBottom() + Math.max(Math.max(reachedHeight, unreachedHeight), Math.abs(textHeight));
            if (mode == MeasureSpec.AT_MOST) {
                result = Math.min(result, size);
            }
        }
        return result;
    }

    private int dp2px(int dpValue) {
        return (int) TypedValue.applyDimension(TypedValue.COMPLEX_UNIT_DIP, dpValue, getResources().getDisplayMetrics());
    }

    private int sp2px(int spValue) {
        return (int) TypedValue.applyDimension(TypedValue.COMPLEX_UNIT_SP, spValue, getResources().getDisplayMetrics());
    }
}
